import io
from types import TracebackType
import requests


class SharepointFile(io.BytesIO):
    """
    SharepointFile(header, sharepoint_site, filepath, checkout)

    A file-like class used to represent a SharePoint file using the SharePoint REST API v1.

    Methods
    -------
    check_back_in()
        Manually check a file into SharePoint and remove the file lock
    """

    def __init__(self, header, sharepoint_site: str, filepath: str, checkout: bool):
        self._header = header
        self._site = sharepoint_site
        self._path, self._filename = self._get_filepath(filepath)
        self._checkout = checkout
        super().__init__(self._get_file())

    def _get_filepath(str, path: str):
        path = path.split("/")
        return "/".join(path[:-1]), path[-1]

    def _get_file(self):
        if self._checkout:
            api_url = f"{self._site}/_api/web/GetFolderByServerRelativeUrl('{self._path}')/Files('{self._filename}')/CheckOut()"
            requests.post(api_url, headers=self._header)

        api_url = f"{self._site}/_api/web/GetFolderByServerRelativeUrl('{self._path}')/Files('{self._filename}')/$value"
        file = requests.get(api_url, headers=self._header)
        return file.content

    def flush(self):
        super().flush()

        api_url = f"{self._site}/_api/web/GetFolderByServerRelativeUrl('{self._path}')/Files/add(url='{self._filename}', overwrite=true)"

        file_size = len(super().getvalue())
        post_request = {
            "Content-Length": str(file_size),
            "X-HTTP-Method": "PUT",
        }

        post_request.update(self._header)
        file_content = super().getvalue()
        requests.put(api_url, data=file_content, headers=post_request)

    def check_back_in(self):
        if self._checkout:
            api_url = f"{self._site}/_api/web/GetFolderByServerRelativeUrl('{self._path}')/Files('{self._filename}')/CheckIn(comment='Comment',checkintype=0)"
            requests.post(api_url, headers=self._header)

    def close(self) -> None:
        self.check_back_in()
        return super().close()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.check_back_in()
        return super().__exit__(exc_type, exc_val, exc_tb)



import json
import requests
from . import sharepoint_file


class SharepointSite:
    """
    SharePointSite(sharepoint_site_name, sharepoint_site_url, sharepoint_base_url, header)

    A class used to represent a SharePoint site using the SharePoint REST API v1.
    ...

    Parameters
    ----------
    sharepoint_site_url : str
        The URL of the Sharepoint site
    sharepoint_base_url : str
        The URL of the Sharepoint instance
    header : dict
        The header for requests made to the API containing the Bearer token

    Attributes
    ----------
    site_name : str
        The user-facing name of the Sharepoint site
    description : str
        The description of the Sharepoint site
    subsites : dict
        A list of dicts of Sharepoint subsites

    Methods
    -------
    listdir(path)
        Returns a list of directories in a document library
    mkdir(path)
        Creates a new directory in a document library
    open(path, checkout = False)
        Downloads a file from a document library and returns a SharepointFile object
    get_subsite(site_name)
        Returns a SharepointSite object for a subsite
    """

    def __init__(
        self,
        sharepoint_site_url: str,
        sharepoint_base_url: str,
        header,
    ) -> None:
        self._site_url = sharepoint_site_url
        self._base_url = sharepoint_base_url
        self._header = header
        self._libraries = self._get_libraries()

    def _get_libraries(self):
        api_url = f"{self._site_url}/_api/web/lists?$select=Title,ServerRelativeUrl&$filter=BaseTemplate eq 101 and hidden eq false&$expand=RootFolder"
        request = requests.get(api_url, headers=self._header)
        request = json.loads(request.content)

        request = request["d"]["results"]
        libraries = {}

        for library in request:
            library_name = library["RootFolder"]["Name"]
            library_path = library["RootFolder"]["ServerRelativeUrl"]
            libraries[library_name] = library_path

        return libraries

    def listdir(self, path: str):
        """List all files in a directory on a SharePoint document library.

        Parameters
        ----------
        path : str
            The path of the directory to search, relative to the site as a whole. File paths are UNIX-like.
        Raises
        ------
        FileNotFoundError
            If the document library does not exist or if a nonexistent folder is searched.
        Exception
            If an API error has occured that is not otherwise caught.
        """

        path_list = path.split("/")
        path_list = list(filter(None, path_list))

        if path_list[0] not in self._libraries:
            raise FileNotFoundError(f"Document Library {path_list[0]} Not Found.")

        api_url = f"{self._site_url}/_api/web/GetFolderByServerRelativeUrl('{'/'.join(path_list)}')?$expand=Folders,Files"
        request = requests.get(api_url, headers=self._header)
        request = json.loads(request.content)

        if "error" in request:
            error_code = request["error"]["code"]
            if error_code == "-2147024894, System.IO.FileNotFoundException":
                raise FileNotFoundError(request["error"]["message"]["value"])
            else:
                raise Exception(request["error"]["message"]["value"])

        request = request["d"]
        files = request["Files"]["results"]
        folders = request["Folders"]["results"]
        directory_list = []

        for file in files:
            directory_list.append(file["Name"])
        for folder in folders:
            directory_list.append(folder["Name"])

        return directory_list

    def mkdir(self, path: str):
        """Create a new directory on a SharePoint document library.

        Parameters
        ----------
        path : str
            The path of the directory to create, relative to the site as a whole. File paths are UNIX-like.
        Raises
        ------
        FileExistsError
            If the folder already exists.
        FileNotFoundError
            If the document library does not exist or if a subfolder is attempted to be made in a nonexistent folder.
        Exception
            If an API error has occured that is not otherwise caught.
        """

        path_list = path.split("/")
        path_list = list(filter(None, path_list))

        if path_list[0] not in self._libraries:
            raise FileNotFoundError(f"Document Library {path_list[0]} not found.")

        if path_list[-1] in self.listdir("/".join(path_list[:-1])):
            raise FileExistsError(f"Folder {path_list[-1]} exists.")

        api_url = f"{self._site_url}/_api/web/folders"
        payload = json.dumps(
            {
                "__metadata": {"type": "SP.Folder"},
                "ServerRelativeUrl": "/".join(path_list),
            }
        )

        request = requests.post(url=api_url, data=payload, headers=self._header)
        request = json.loads(request.content)

        if "error" in request:
            error_code = request["error"]["code"]
            if error_code == "-2130247139, Microsoft.SharePoint.SPException":
                raise FileNotFoundError(request["error"]["message"]["value"])
            else:
                raise Exception(request["error"]["message"]["value"])

    def open(self, filepath: str, checkout: bool = False):
        """Open a file from a SharePoint document library and return a file-like object.

        Parameters
        ----------
        filepath : str
            The path of the file to return, relative to the site as a whole. File paths are UNIX-like.
        checkout : bool
            If True, the file will be checked out of Sharepoint and locked.
        Returns
        ------
        SharepointFile
            File-like object.
        """

        return sharepoint_file.SharepointFile(
            header=self._header,
            sharepoint_site=self._site_url,
            filepath=filepath,
            checkout=checkout,
        )

    @property
    def site_name(self):
        api_url = f"{self._site_url}/_api/web/title"
        request = json.loads(requests.get(api_url, headers=self._header).content)
        return request["d"]["Title"]

    @property
    def description(self):
        api_url = f"{self._site_url}/_api/web/description"
        request = json.loads(requests.get(api_url, headers=self._header).content)
        return request["d"]["Description"]

    @property
    def subsites(self):
        api_url = f"{self._site_url}/_api/web/webs/?$select=title,Url"
        request = requests.get(api_url, headers=self._header).text
        request = json.loads(request)
        request = request["d"]["results"]
        sites = []
        for x in request:
            site_dict = {
                "Site Name": x["Title"],
                "Site Path": x["Url"],
            }
            sites.append(site_dict)
        return sites

    def get_subsite(self, site_name: str):
        """Grab a subsite,

        Parameters
        ----------
        site_name : str
            The user-facing name of a SharePoint subsite

        Raises
        ------
        KeyError
            If the subsite does not exist.
        """

        site_url = next(
            (item for item in self.sites if item["Site Name"] == site_name), None
        )
        if site_url is None:
            raise KeyError("Site not found.")
        else:
            site_url = site_url["Site Path"]
        return SharepointSite(site_url, self.base_url, self._header)



import azure.identity
import requests
import json
from . import sharepoint_site
import azure.core.credentials


class SharePoint:
    """
    A class used to represent an organization's SharePoint instance using the SharePoint REST API v1.
    ...

    Parameters
    ----------
    base_url : str
        The URL of a Sharepoint instance, ex. contoso.sharepoint.com
    azure_identity : TokenCredential
        An azure-identity token credential.

    Attributes
    ----------
    sites : dict
        a dictionary of all sites in SharePoint, the key is the user-facing name and the value is the URL
    base_url : str
        the URL of the SharePoint instance

    Methods
    -------
    get_site(site_name)
        Returns a SharepointSite object for a specific SharePoint site
    """

    def __init__(
        self,
        sharepoint_url: str,
        azure_identity: azure.core.credentials.TokenCredential,
    ) -> None:
        self.base_url = sharepoint_url
        self._scope = f"{self.base_url}/.default"
        self._identity = azure_identity
        self.sites = self._initalize_sites()

    @property
    def _token(self):
        return self._identity.get_token(self._scope)

    @property
    def _header(self):
        return {
            "Authorization": f"Bearer {self._token.token}",
            "Accept": "application/json;odata=verbose",
            "Content-Type": "application/json;odata=verbose",
        }

    def _initalize_sites(self):
        api_url = f"{self.base_url}/_api/search/query?querytext='contentclass:STS_Site contentclass:STS_Web'&selectproperties='Title,Path'"
        request = requests.get(api_url, headers=self._header).text
        request = json.loads(request)
        # fmt: off
        request = request["d"]["query"]["PrimaryQueryResult"]["RelevantResults"]["Table"]["Rows"]["results"]
        # fmt: on

        sites = []
        for x in request:
            site_dict = {
                "Site Name": x["Cells"]["results"][0]["Value"],
                "Site Path": x["Cells"]["results"][1]["Value"],
            }
            sites.append(site_dict)
        return sites

    def get_site(self, site_name):
        """
        Parameters
        ----------
        site_name : str
            The user-facing name of a SharePoint site

        Raises
        ------
        KeyError
            If the subsite does not exist

        """

        site_url = next(
            (item for item in self.sites if item["Site Name"] == site_name), None
        )
        if site_url is None:
            raise KeyError("Site not found.")
        else:
            site_url = site_url["Site Path"]
        return sharepoint_site.SharepointSite(site_url, self.base_url, self._header)



from sharkpoint import sharepoint
from sharkpoint import sharepoint_file
from sharkpoint import sharepoint_site

from sharkpoint.sharepoint import (
    SharePoint,
)
from sharkpoint.sharepoint_file import (
    SharepointFile,
)
from sharkpoint.sharepoint_site import (
    SharepointSite,
)

__all__ = [
    "SharePoint",
    "SharepointFile",
    "SharepointSite",
    "sharepoint",
    "sharepoint_file",
    "sharepoint_site",
]
